from __future__ import annotations

import collections
import dataclasses
import enum
import logging
from pathlib import Path
from typing import Any, Callable, ClassVar, Iterable, Iterator, MutableMapping, Optional, Sequence, TypeVar, overload

from kraken.common import CurrentDirectoryProjectFinder, ProjectFinder, ScriptRunner
from nr.stream import Stream
from typing_extensions import TypeAlias

from kraken.core.address import Address, AddressSpace, resolve_address
from kraken.core.base import Currentable, MetadataContainer
from kraken.core.system.errors import BuildError, ProjectLoaderError, ProjectNotFoundError
from kraken.core.system.executor import GraphExecutor, GraphExecutorObserver
from kraken.core.system.executor.default import (
    DefaultGraphExecutor,
    DefaultPrintingExecutorObserver,
    DefaultTaskExecutor,
)
from kraken.core.system.graph import TaskGraph
from kraken.core.system.project import Project
from kraken.core.system.task import Task

logger = logging.getLogger(__name__)
T = TypeVar("T")


class KrakenAddressSpace(AddressSpace["Project | Task"]):
    """
    Implements the navigation for address resolution in the addressable project and task namespace
    represented by a #Context and its root #Project.
    """

    def __init__(self, root_project: Project) -> None:
        self.root_project = root_project

    def get_root(self) -> Project:
        return self.root_project

    def get_parent(self, entity: Project | Task) -> Project | None:
        if isinstance(entity, Task):
            return entity.project
        return entity.parent

    def get_children(self, entity: Project | Task) -> Iterable[Project | Task]:
        if isinstance(entity, Project):
            yield from entity.subprojects().values()
            yield from entity.tasks().values()


class ContextEventType(enum.Enum):
    any = enum.auto()
    on_project_init = enum.auto()  # event data type is Project
    on_project_loaded = enum.auto()  # event data type is Project
    on_project_begin_finalize = enum.auto()  # event data type is Project
    on_project_finalized = enum.auto()  # event data type is Project
    on_context_begin_finalize = enum.auto()  # event data type is Context
    on_context_finalized = enum.auto()  # event data type is Context


@dataclasses.dataclass
class ContextEvent:
    Type: ClassVar[TypeAlias] = ContextEventType
    Listener = Callable[["ContextEvent"], Any]
    T_Listener = TypeVar("T_Listener", bound=Listener)

    type: Type
    data: Any  # Depends on the event type


class TaskResolutionException(Exception):
    pass


class Context(MetadataContainer, Currentable["Context"]):
    """This class is the single instance where all components of a build process come together."""

    #: The focus project is the one that maps to the current working directory when invoking kraken.
    #: Kraken may be invoked in a directory that does not map to a project, in which case this is None.
    focus_project: Project | None = None

    def __init__(
        self,
        build_directory: Path,
        project_finder: ProjectFinder | None = None,
        executor: GraphExecutor | None = None,
        observer: GraphExecutorObserver | None = None,
    ) -> None:
        """
        :param build_directory: The directory in which all files generated durin the build should be stored.
        :param project_finder: This project finder should only search within the directory it was given, not
            around or in parent folders. Defaults to :class:`CurrentDirectoryProjectFinder`.
        :param executor: The executor to use when the graph is executed.
        :param observer: The executro observer to use when the graph is executed.
        """

        super().__init__()
        self.build_directory = build_directory
        self.project_finder = project_finder or CurrentDirectoryProjectFinder.default()
        self.executor = executor or DefaultGraphExecutor(DefaultTaskExecutor())
        self.observer = observer or DefaultPrintingExecutorObserver()
        self._finalized: bool = False
        self._root_project: Optional[Project] = None
        self._listeners: MutableMapping[ContextEvent.Type, list[ContextEvent.Listener]] = collections.defaultdict(list)
        self.focus_project: Project | None = None

    @property
    def root_project(self) -> Project:
        assert self._root_project is not None, "Context.root_project is not set"
        return self._root_project

    @root_project.setter
    def root_project(self, project: Project) -> None:
        assert self._root_project is None, "Context.root_project is already set"
        self._root_project = project

    def load_project(
        self,
        directory: Path,
        parent: Project | None = None,
        require_buildscript: bool = True,
        runner: ScriptRunner | None = None,
        script: Path | None = None,
    ) -> Project:
        """Loads a project from a file or directory.

        :param directory: The directory to load the project from.
        :param parent: The parent project. If no parent is specified, then the :attr:`root_project`
            must not have been initialized yet and the loaded project will be initialize it.
            If the root project is initialized but no parent is specified, an error will be
            raised.
        :param require_buildscript: If set to `True`, a build script must exist in *directory*.
            Otherwise, it will be accepted if no build script exists in the directory.
        :param runner: If the :class:`ScriptRunner` for this project is already known, it can be passed here.
        :param script: If the script to load for the project is already known, it can be passed here. Cannot be
            specified without a *runner*.
        """

        if not runner:
            if script is not None:
                raise ValueError("cannot specify `script` parameter without a `runner` parameter")
            project_info = self.project_finder.find_project(directory)
            if project_info is not None:
                script, runner = project_info
        if not script and runner:
            script = runner.find_script(directory)

        has_root_project = self._root_project is not None
        project = Project(directory.name, directory, parent, self)
        try:
            if parent:
                parent.add_child(project)

            self.trigger(ContextEvent.Type.on_project_init, project)

            with self.as_current(), project.as_current():
                if not has_root_project:
                    self._root_project = project

                if script is None and require_buildscript:
                    raise ProjectLoaderError(
                        project,
                        f"no buildscript found for {project} (directory: {project.directory.absolute().resolve()})",
                    )
                if script is not None:
                    assert runner is not None
                    runner.execute_script(script, {"project": project})

            self.trigger(ContextEvent.Type.on_project_loaded, project)

        except ProjectLoaderError as exc:
            if exc.project is project:
                # Revert changes if the project that the error occurred with is the current project.
                if not has_root_project:
                    self._root_project = None
                if parent:
                    parent.remove_child(project)
            raise

        return project

    def iter_projects(self, relative_to: Project | None = None) -> Iterator[Project]:
        """Iterates over all projects in the context."""

        def _recurse(project: Project) -> Iterator[Project]:
            yield project
            for child_project in project.subprojects().values():
                yield from _recurse(child_project)

        yield from _recurse(relative_to or self.root_project)

    def get_project(self, address: Address) -> Project:
        """
        Find a project by its address. The address must be absolute.
        """

        if not address.is_absolute():
            raise ValueError(f"address '{address}' is not absolute")

        project: Project | None = self.root_project
        assert project is not None

        for element in address.elements:
            project = project.subproject(element.value, load=False)
            if not project:
                raise ProjectNotFoundError(address)

        return project

    def resolve_tasks(
        self,
        addresses: Sequence[str | Address] | None,
        relative_to: Project | Address | None = None,
        set_selected: bool = False,
    ) -> list[Task]:
        """
        This method finds Kraken tasks by their address, relative to a given project. If no project is
        specified, the address is resolved relative to the root project.

        :param addresses: A list of task addresses to resolve. Task addresses may contain glob patterns
            (`*` and `**` as well as `?` at the end of an address element, see the #Address class for
            more details).

            Any address that consists of only a single non-globbing path element (such as `lint` or `test`)
            will be prefixed by a wildcard (such that they are semantically equivalent to `**:lint` and
            `**:test`, respectively).

            In case the address specifies a container (that is, if it ends with a colon), then this will
            resolve the default tasks or this container.
            As an example, `:` will get the default tasks of the current project, and `:**:` will get the
            default tasks of all sub-projects.
            Specifying `None` is a shorthand for resolving `:` and `:**:`, that is, will resolve to the
            default tasks of the current project and its sub-projects.

        :param relative_to: The Kraken project to resolve the task addresses relative to. If this is not
            specified, the #root_project is used instead.

        :param set_selected: If enabled, addresses that resolve to tasks immediately will be marked as selected
            before they are returned. Note that this does not mark tasks as selected when they are picked up by
            via the default tasks of a project. For example, when `:*` is resolved, the default tasks of all
            sub-projects will be returned, but they will not be marked as selected. The tasks of the root project
            however, will be marked as selected.
        """

        if not isinstance(relative_to, Address):
            relative_to = relative_to.address if relative_to is not None else Address.ROOT

        if not relative_to.is_absolute():
            raise ValueError(f"'relative_to' must be an absolute address (got {relative_to!r})")

        if addresses is None:
            addresses = [
                ".",  # The current project (will be "expanded" to its default tasks)
                "**:",  # All sub-projects (will be "expanded" to their default tasks)
            ]

        results: list[Task] = []
        space = KrakenAddressSpace(self.root_project)
        for address in addresses:
            try:
                results += self._resolve_single_address(Address(address), relative_to, space, set_selected)
            except TaskResolutionException:
                if address == "**:":
                    # In case the project has no sub-projects, it is expected not to find any tasks there
                    pass
                else:
                    raise

        return results

    def _resolve_single_address(
        self,
        address: Address,
        relative_to: Address,
        space: KrakenAddressSpace,
        set_selected: bool,
    ) -> list[Task]:
        """
        Resolve a single address in the context.

        Any address that contains only a single path element (such as `lint` or `test`) will be prefixed
        with `**:`, such that they are semantically equivalent to `**:lint` and `**:test`, respectively.
        """

        if address.is_empty():
            raise TaskResolutionException("Impossible to resolve the empty address.")

        # Prefix single-element addresses with `**:`, unless the last element already is `**`.
        if not address.is_absolute() and len(address) == 1 and not address.elements[0].is_recursive_wildcard():
            address = Address.RECURSIVE_WILDCARD.concat(address)
        if not address.is_absolute():
            address = relative_to.concat(address)

        matches = list(resolve_address(space, self.root_project, address).matches())
        tasks = Stream(matches).of_type(Task).collect()  # type: ignore[type-abstract]
        if set_selected:
            for task in tasks:
                task.selected = True
        projects = Stream(matches).of_type(Project).collect()
        if projects:
            # Using the address of a project means we want to select its default tasks
            for proj in projects:
                tasks += [task for task in proj.tasks().values() if task.default]
        if not tasks:
            raise TaskResolutionException(f"'{address}' refers to no tasks.")
        return tasks

    def finalize(self) -> None:
        """Call :meth:`Task.finalize()` on all tasks. This should be called before a graph is created."""

        if self._finalized:
            logger.warning("Context.finalize() called more than once", stack_info=True)
            return

        self._finalized = True
        self.trigger(ContextEvent.Type.on_context_begin_finalize, self)

        # Delegate to finalize calls in all tasks of all projects.
        for project in self.iter_projects():
            self.trigger(ContextEvent.Type.on_project_begin_finalize, project)
            for task in project.tasks().values():
                task.finalize()
            self.trigger(ContextEvent.Type.on_project_finalized, project)

        self.trigger(ContextEvent.Type.on_context_finalized, self)

    def get_build_graph(self, targets: Sequence[str | Task] | None) -> TaskGraph:
        """Returns the :class:`TaskGraph` that contains either all default tasks or the tasks specified with
        the *targets* argument.

        :param targets: A list of targets to resolve and to build the graph from.
        :raise ValueError: If not tasks were selected.
        """

        if targets is None:
            tasks = self.resolve_tasks(None)
        else:
            tasks = self.resolve_tasks([t for t in targets if isinstance(t, str)]) + [
                t for t in targets if not isinstance(t, str)
            ]

        if not tasks:
            raise ValueError("no tasks selected")

        graph = TaskGraph(self).trim(tasks)

        assert graph, "TaskGraph cannot be empty"
        return graph

    def execute(self, tasks: list[str | Task] | TaskGraph | None = None) -> None:
        """Execute all default tasks or the tasks specified by *targets* using the default executor.
        If :meth:`finalize` was not called already it will be called by this function before the build
        graph is created, unless a build graph is passed in the first place.

        :param tasks: The list of tasks to execute, or the build graph. If none specified, all default
            tasks will be executed.
        :raise BuildError: If any task fails to execute.
        """

        if isinstance(tasks, TaskGraph):
            assert self._finalized, "no, no, this is all wrong. you need to finalize the context first"
            graph = tasks
        else:
            if not self._finalized:
                self.finalize()
            graph = self.get_build_graph(tasks)

        self.executor.execute_graph(graph, self.observer)

        if not graph.is_complete():
            failed_tasks = list(graph.tasks(failed=True))
            if len(failed_tasks) == 1:
                message = f'task "{failed_tasks[0].path}" failed'
            else:
                message = "tasks " + ", ".join(f'"{task.path}"' for task in failed_tasks) + " failed"
            raise BuildError(message)

    @overload
    def listen(
        self, event_type: str | ContextEvent.Type
    ) -> Callable[[ContextEvent.T_Listener], ContextEvent.T_Listener]:
        ...

    @overload
    def listen(self, event_type: str | ContextEvent.Type, listener: ContextEvent.Listener) -> None:
        ...

    def listen(self, event_type: str | ContextEvent.Type, listener: ContextEvent.Listener | None = None) -> Any:
        """Registers a listener to the context for the given event type."""

        if isinstance(event_type, str):
            event_type = ContextEvent.Type[event_type]

        def register(listener: ContextEvent.T_Listener) -> ContextEvent.T_Listener:
            assert callable(listener), "listener must be callable, got: %r" % listener
            self._listeners[event_type].append(listener)
            return listener

        if listener is None:
            return register

        register(listener)

    def trigger(self, event_type: ContextEvent.Type, data: Any) -> None:
        assert isinstance(event_type, ContextEvent.Type), repr(event_type)
        assert event_type != ContextEvent.Type.any, "cannot trigger event of type 'any'"
        listeners = (*self._listeners.get(ContextEvent.Type.any, ()), *self._listeners.get(event_type, ()))
        for listener in listeners:
            # TODO(NiklasRosenstein): Should we catch errors in listeners of letting them propagate?
            listener(ContextEvent(event_type, data))



"""
Provides the :class:`Currentable` base class which outfits subclasses with an :meth:`~Currentable.as_current` method
that makes a single instance of a class available globally (across threads, but not thread-safe).
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Any, ClassVar, Generic, Iterator, TypeVar, cast, overload

from kraken.common import NotSet

T = TypeVar("T")
U = TypeVar("U")


class CurrentProvider(ABC, Generic[T]):
    """
    Protocol for classes that provide a :meth:`current` classmethod.
    """

    @overload
    @classmethod
    def current(cls) -> T:
        """Returns the current context or raises a :class:`RuntimeError`."""

    @overload
    @classmethod
    def current(cls, fallback: U) -> T | U:
        """Returns the current context or *fallback*."""

    @classmethod
    def current(cls, fallback: U | NotSet = NotSet.Value) -> T | U:
        try:
            return cls._get_current_object()
        except RuntimeError:
            if fallback is NotSet.Value:
                raise
            return fallback

    @classmethod
    @abstractmethod
    def _get_current_object(cls) -> T:
        raise NotImplementedError


class Currentable(CurrentProvider[T]):
    """
    Base class for classes that should have a :meth:`as_current` method.
    """

    __current: ClassVar[Any | None] = None  # note: ClassVar cannot contain type variables

    @contextmanager
    def as_current(self) -> Iterator[None]:
        """
        A context manager that makes the instance *self* available globally to be retrieved with :meth:`current`.

        This method is not thread-safe, and the object will be available for all threads.
        """

        prev = type(self).__current
        try:
            type(self).__current = self
            yield
        finally:
            type(self).__current = prev

    # CurrentProvider

    @classmethod
    def _get_current_object(cls) -> T:
        if cls.__current is None:
            raise RuntimeError(f"No current object for type `{cls.__name__}`")
        return cast(T, cls.__current)



from __future__ import annotations

import contextlib
import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Iterator

import pytest

from kraken.core.system.context import Context
from kraken.core.system.project import Project

if TYPE_CHECKING:
    from _pytest.fixtures import FixtureRequest

__all__ = [
    "kraken_ctx",
    "kraken_project",
]

logger = logging.getLogger(__name__)


@pytest.fixture(name="kraken_ctx")
def _kraken_ctx_fixture() -> Iterator[Context]:
    with kraken_ctx() as ctx:
        yield ctx


@contextlib.contextmanager
def kraken_ctx() -> Iterator[Context]:
    context = Context(Path("build"))
    with context.as_current():
        yield context


@pytest.fixture(name="kraken_project")
def _kraken_project_fixture(kraken_ctx: Context, request: FixtureRequest) -> Iterator[Project]:
    with kraken_project(kraken_ctx, request.path) as project:
        yield project


@contextlib.contextmanager
def kraken_project(kraken_ctx: Context, path: Path | None = None) -> Iterator[Project]:
    if path is None:
        path = Path(sys._getframe(1).f_code.co_filename)
    kraken_ctx.root_project = Project("test", path.parent, None, kraken_ctx)
    with kraken_ctx.root_project.as_current():
        yield kraken_ctx.root_project



from __future__ import annotations

from typing import Any, Callable, List, TypeVar, overload

T = TypeVar("T")


class MetadataContainer:
    """
    A base class for classes that shall have a :attr:`metadata` attribute that is intended to contain arbitrary
    objects to store metadata on the object. The metadata can be retrieved later using the :meth:`find_metadata`
    method.
    """

    metadata: List[Any]

    def __init__(self) -> None:
        self.metadata = []

    @overload
    def find_metadata(self, of_type: type[T]) -> T | None:
        """Returns the first entry in the :attr:`metadata` that is of the specified type."""

    @overload
    def find_metadata(self, of_type: type[T], create: Callable[[], T]) -> T:
        """Returns the first entry in :attr:`metadata`, or creates one."""

    def find_metadata(self, of_type: type[T], create: Callable[[], T] | None = None) -> T | None:
        obj = next((x for x in self.metadata if isinstance(x, of_type)), None)
        if obj is None and create is not None:
            obj = create()
            self.metadata.append(obj)
        return obj



from dataclasses import dataclass

import pytest

from kraken.core.system.project import Project
from kraken.core.system.property import Property
from kraken.core.system.task import Task, VoidTask


@dataclass
class MyDescriptor:
    name: str


def test__Project__resolve_outputs__can_find_dataclass_in_metadata(kraken_project: Project) -> None:
    kraken_project.do("carrier", VoidTask).outputs.append(MyDescriptor("foobar"))
    assert list(kraken_project.resolve_tasks(":carrier").select(MyDescriptor).all()) == [MyDescriptor("foobar")]


def test__Project__resolve_outputs__can_find_dataclass_in_properties(kraken_project: Project) -> None:
    class MyTask(Task):
        out_prop: Property[MyDescriptor] = Property.output()

        def execute(self) -> None:
            ...

    kraken_project.do("carrier", MyTask, out_prop=MyDescriptor("foobar"))
    assert list(kraken_project.resolve_tasks(":carrier").select(MyDescriptor).all()) == [MyDescriptor("foobar")]


def test__Project__resolve_outputs__can_not_find_input_property(kraken_project: Project) -> None:
    class MyTask(Task):
        out_prop: Property[MyDescriptor]

        def execute(self) -> None:
            ...

    kraken_project.do("carrier", MyTask, out_prop=MyDescriptor("foobar"))
    assert list(kraken_project.resolve_tasks(":carrier").select(MyDescriptor).all()) == []


def test__Project__resolve_outputs_supplier(kraken_project: Project) -> None:
    class MyTask(Task):
        out_prop: Property[MyDescriptor] = Property.output()

        def execute(self) -> None:
            ...

    kraken_project.do("carrier", MyTask, out_prop=MyDescriptor("foobar"))
    assert kraken_project.resolve_tasks(":carrier").select(MyDescriptor).supplier().get() == [MyDescriptor("foobar")]


def test__Project__do_normalizes_taskname_backwards_compatibility_pre_0_12_0(kraken_project: Project) -> None:
    with pytest.warns(DeprecationWarning) as warninfo:
        task = kraken_project.do("this is a :test task", VoidTask)
    assert task.name == "this-is-a-test-task"
    assert str(warninfo.list[0].message) == (
        "Task name `this is a :test task` is invalid and will be normalized to `this-is-a-test-task`. "
        "Starting with kraken-core 0.12.0, Task names must follow a stricter naming convention subject to the "
        "Address class' validation (must match /^[a-zA-Z0-9/_\\-\\.\\*]+$/)."
    )


def test__Project__do__does_not_set_property_on_None_value(kraken_project: Project) -> None:
    class MyTask(Task):
        in_prop: Property[str]

        def execute(self) -> None:
            ...

    kraken_project.do("carrier", MyTask, in_prop=None)
    assert kraken_project.resolve_tasks(":carrier").select(str).supplier().get() == []



from pathlib import Path

from pytest import raises

from kraken.core.address import AddressResolutionError
from kraken.core.system.context import Context, TaskResolutionException
from kraken.core.system.project import Project
from kraken.core.system.task import VoidTask


def test__Context__resolve_tasks(kraken_ctx: Context, kraken_project: Project) -> None:
    # Generate tasks
    kraken_project.do("default1", VoidTask, default=True)
    kraken_project.do("default2", VoidTask, default=True)
    kraken_project.do("task_from_group_g1", VoidTask, group="g1")
    kraken_project.do("other_task_from_group_g1", VoidTask, group="g1")
    kraken_project.do("no_group", VoidTask)

    sub_proj = Project("sub_proj", Path(), parent=kraken_project, context=kraken_ctx)
    kraken_project.add_child(sub_proj)
    sub_proj.do("subproject_task", VoidTask, default=True)
    sub_proj.do("subproject_task_in_group", VoidTask, group="g1")
    sub_proj.do("subproject_task2", VoidTask)

    # Search for tasks from their addresses
    tasks = kraken_ctx.resolve_tasks(["default1"])
    assert len(tasks) == 1

    tasks = kraken_ctx.resolve_tasks(["subproject_task"])
    assert len(tasks) == 1

    with raises(AddressResolutionError):
        kraken_ctx.resolve_tasks([":subproject_task"])

    tasks = kraken_ctx.resolve_tasks(["g1"])
    assert len(tasks) == 2

    tasks = kraken_ctx.resolve_tasks([":g1"])
    assert len(tasks) == 1

    tasks = kraken_ctx.resolve_tasks(["defa*"])
    assert len(tasks) == 2

    # Search for default tasks...
    # ...in the current project
    default_tasks = kraken_ctx.resolve_tasks([":"])
    default_voidtasks = [task for task in default_tasks if type(task) == VoidTask]
    assert len(default_voidtasks) == 2
    assert default_voidtasks[0].name == "default1"
    assert default_voidtasks[1].name == "default2"

    # ...and in subprojects only
    default_tasks = kraken_ctx.resolve_tasks([":**:"])
    default_voidtasks = [task for task in default_tasks if type(task) == VoidTask]
    assert len(default_voidtasks) == 1
    assert default_voidtasks[0].name == "subproject_task"

    with raises(TaskResolutionException):
        kraken_ctx.resolve_tasks([""])

    nothing = kraken_ctx.resolve_tasks([])
    assert len(nothing) == 0


def test__Context__resolve_tasks__can_resolve_optional_tasks(kraken_ctx: Context, kraken_project: Project) -> None:
    """
    Tests that optional tasks can be resolved using the #Context.resolve_tasks() method.
    """

    sub1 = kraken_project.subproject("sub1", mode="empty")
    sub2 = kraken_project.subproject("sub2", mode="empty")

    publish_task = sub2.do("publishSomething", VoidTask, group="publish")
    assert sub2.group("publish").tasks == [sub2.task("publishSomething")]

    # Note: A "publish" task group already exists on all projects by default.

    # When we resolve the address selector "publish", it matches all tasks named "publish" in all projects.
    tasks = kraken_ctx.resolve_tasks(["publish"])
    assert set(tasks) == {
        kraken_project.group("publish"),
        sub1.group("publish"),
        sub2.group("publish"),
    }

    # When resolving it inside a sub project, it matches only the tasks in that sub project.
    tasks = kraken_ctx.resolve_tasks(["publish"], relative_to=sub2)
    assert set(tasks) == {sub2.group("publish")}

    #
    # Let's also test the tasks that the TaskGraph will contain.
    #

    # When we pass all publish tasks, they are definitely included in the graph + the members of those groups.
    expanded_tasks = set(kraken_ctx.get_build_graph(kraken_ctx.resolve_tasks(["publish"])).tasks())
    assert set(expanded_tasks) == {
        kraken_project.group("publish"),
        sub1.group("publish"),
        sub2.group("publish"),
        publish_task,
    }

    # When we pass the root publish task, only that group is in the graph because that group has no members.
    expanded_tasks = set(kraken_ctx.get_build_graph([kraken_project.task("publish")]).tasks())
    assert set(expanded_tasks) == {kraken_project.group("publish")}

    # When we pass the publish task from the sub project that contains a task, those are in the graph.
    expanded_tasks = set(kraken_ctx.get_build_graph([sub2.task("publish")]).tasks())
    assert set(expanded_tasks) == {sub2.group("publish"), publish_task}



from __future__ import annotations

import argparse
import builtins
import contextlib
import io
import json
import logging
import os
import pdb
import sys
import textwrap
from functools import partial
from itertools import chain
from pathlib import Path
from typing import Any, NoReturn

import pretty_errors  # type: ignore
from kraken.common import (
    BuildscriptMetadata,
    CurrentDirectoryProjectFinder,
    LoggingOptions,
    RequirementSpec,
    appending_to_sys_path,
    deprecated_get_requirement_spec_from_file_header,
    get_terminal_width,
    not_none,
    propagate_argparse_formatter_to_subparser,
)
from kraken.common.pyenv import get_distributions
from nr.io.graphviz.render import render_to_browser
from nr.io.graphviz.writer import GraphvizWriter
from termcolor import colored

from kraken.core import __version__
from kraken.core.address import Address
from kraken.core.cli import serialize
from kraken.core.cli.executor import ColoredDefaultPrintingExecutorObserver, status_to_text
from kraken.core.cli.option_sets import BuildOptions, ExcludeOptions, GraphOptions, RunOptions, VizOptions
from kraken.core.system.context import Context
from kraken.core.system.errors import BuildError, ProjectNotFoundError
from kraken.core.system.graph import TaskGraph
from kraken.core.system.project import Project
from kraken.core.system.property import Property
from kraken.core.system.task import GroupTask, Task, TaskStatus

pretty_errors.configure(
    separator_character=pretty_errors.CYAN + "≈",
    filename_display=pretty_errors.FILENAME_COMPACT,
    line_number_first=False,
    lines_before=2,
    lines_after=2,
    display_link=True,
    display_locals=False,
    display_arrow=True,
    display_trace_locals=False,
    display_timestamp=False,
    prefix=pretty_errors.MAGENTA
    + "\nLet no joyful voice be heard! Let no man look up at the sky with hope!"
    + " And let this day be cursed by we who ready to wake... the Kraken!"
    + pretty_errors.default_config.line_color
    + "\nAlas sailor, there is nought you can do. Please check your Kraken build"
    + " script, and then report this stacktrace to the Kraken repository:\n"
    + pretty_errors.CYAN
    + "https://github.com/kraken-build/kraken-build/issues\n\n"
    + pretty_errors.MAGENTA
    + "> "
    + pretty_errors.default_config.line_color
    + "Kraken Version: "
    + pretty_errors.BRIGHT_MAGENTA
    + __version__
    + pretty_errors.MAGENTA
    + "\n>",
    infix=pretty_errors.MAGENTA + "> ",
    arrow_head_character="  ↑",
    arrow_tail_character=" ",
    trace_lines_before=0,
    trace_lines_after=0,
    header_color=pretty_errors.default_config.line_color,
    line_color=pretty_errors.MAGENTA + "> " + pretty_errors.default_config.line_color,
    link_color=pretty_errors.CYAN + "> " + pretty_errors.default_config.line_color,
    code_color=pretty_errors.CYAN + "> " + pretty_errors.default_config.line_color,
    line_number_color=pretty_errors.MAGENTA + "#" + pretty_errors.default_config.line_color,
    filename_color=pretty_errors.MAGENTA + "> " + pretty_errors.default_config.line_color,
    exception_color=pretty_errors.CYAN + "> " + pretty_errors.YELLOW,
    exception_arg_color=pretty_errors.MAGENTA + "> " + pretty_errors.YELLOW,
    exception_file_color=pretty_errors.MAGENTA + "> " + pretty_errors.YELLOW,
    function_color=pretty_errors.CYAN,
    local_name_color=pretty_errors.CYAN + "> " + pretty_errors.default_config.line_color,
    local_value_color=pretty_errors.YELLOW,
    inner_exception_message=pretty_errors.MAGENTA
    + "\n> "
    + pretty_errors.YELLOW
    + "WARNING: "
    + pretty_errors.default_config.line_color
    + "The above exception was the direct cause of the following exception:\n",
    syntax_error_color=pretty_errors.MAGENTA,
    arrow_head_color=pretty_errors.CYAN,
    inner_exception_separator=True,
    show_suppressed=True,
)

if not pretty_errors.terminal_is_interactive:
    pretty_errors.mono()


BUILD_SCRIPT = Path(".kraken.py")
BUILD_SUPPORT_DIRECTORY = "build-support"
logger = logging.getLogger(__name__)
print = partial(builtins.print, flush=True)


def _get_argument_parser(prog: str) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog,
        formatter_class=lambda prog: argparse.RawDescriptionHelpFormatter(prog, width=120, max_help_position=60),
        description=textwrap.dedent(
            """
            The Kraken build system.

            Kraken focuses on ease of use and simplicity to model complex task orchestration workflows.
            """
        ),
    )
    subparsers = parser.add_subparsers(dest="cmd")

    run = subparsers.add_parser(
        "run",
        aliases=["r"],
        # The following makes sure --all will not be considered as an abbreviation of --allow-whatever
        # This is important because we do want to warn in case the user specifies the (invalid) --all flag
        # See https://docs.python.org/dev/library/argparse.html#allow-abbrev
        allow_abbrev=False,
    )
    LoggingOptions.add_to_parser(run)
    BuildOptions.add_to_parser(run)
    GraphOptions.add_to_parser(run, include_all_option=False)
    RunOptions.add_to_parser(run)

    query = subparsers.add_parser("query", aliases=["q"])
    query_subparsers = query.add_subparsers(dest="query_cmd")

    ls = query_subparsers.add_parser("ls", description="list all tasks and task groups in the build")
    LoggingOptions.add_to_parser(ls)
    BuildOptions.add_to_parser(ls)
    GraphOptions.add_to_parser(ls, saveable=False)

    describe = query_subparsers.add_parser(
        "describe",
        aliases=["d"],
        description="describe one or more tasks in detail",
    )
    LoggingOptions.add_to_parser(describe)
    BuildOptions.add_to_parser(describe)
    GraphOptions.add_to_parser(describe, saveable=False)

    viz = query_subparsers.add_parser("visualize", aliases=["viz", "v"], description="generate a GraphViz of the build")
    LoggingOptions.add_to_parser(viz)
    BuildOptions.add_to_parser(viz)
    GraphOptions.add_to_parser(viz, saveable=False)
    VizOptions.add_to_parser(viz)

    tree = query_subparsers.add_parser("tree", aliases=["t"], description="Output the project and task tree.")
    LoggingOptions.add_to_parser(tree)
    BuildOptions.add_to_parser(tree)
    GraphOptions.add_to_parser(tree)
    ExcludeOptions.add_to_parser(tree)

    # This command is used by kraken-wrapper to produce a lock file.
    env = query_subparsers.add_parser("env", description="produce a JSON file of the Python environment distributions")
    LoggingOptions.add_to_parser(env)

    propagate_argparse_formatter_to_subparser(parser)
    return parser


def _load_build_state(
    exit_stack: contextlib.ExitStack,
    build_options: BuildOptions,
    graph_options: GraphOptions,
) -> tuple[Context, TaskGraph]:
    """
    This function loads the build state for the current working directory; which involves either executing the
    Kraken build script or loading one or more state files from their serialized form on disk.
    """

    if graph_options.restart and not graph_options.resume:
        raise ValueError("the --restart option requires the --resume flag")

    # Calculate the main subproject based on the project directory.
    root_directory = build_options.project_dir.absolute().resolve()
    try:
        subproject_directory = Path.cwd().relative_to(root_directory)
    except ValueError:
        raise ValueError(
            f"-p,--project-dir must be a parent directory of {Path.cwd()}, not a sibling/subdirectory "
            f"(got: {build_options.project_dir})"
        )

    # For consistency, we always act as if Kraken was run from the project root directory.
    # Using the `subproject_directory`, we later fitler down which tasks are selected / how relative
    # task references on the CLI are resolved.
    os.chdir(root_directory)

    project_info = CurrentDirectoryProjectFinder.default().find_project(Path.cwd())
    if not project_info:
        # We are OKAY with resuming a build from serialized state files even if no build script exists in the
        # current working directory; this is a feature that is often useful for debugging purposes when you want
        # to inspect the final state of a build, like from CI.
        if not graph_options.resume:
            raise ValueError(f'no Kraken build script found in the directory "{build_options.project_dir}"')

    # Before we can deserialize the build state, we must add the additional paths to `sys.path` that are defined
    # in by the script using the buildscript() function, or for backwards compatibility, in the file header as
    # comments.

    # Note that if we are simply going to execute the build script (i.e. not deserializing from state files),
    # we can rely on the buildscript() call in the script to update `sys.path`; but if the deprecated file header
    # is used to define the pythonpath we still need to parse it explicitly.

    if project_info:
        # Attempt to read the requirement spec in the deprecated format first.
        requirements = deprecated_get_requirement_spec_from_file_header(project_info.script)

        # If the file does not have the deprecated requirement spec file header as comments, we instead want
        # to capture the buildscript() call by tenatively executing the script. However, we only need to do
        # this if we want to resume from a serialized build state. When we need to execute the full script
        # anyway, we can rely on a callback that we register for when buildscript() is called to update
        # the `sys.path`, which avoids that we execute the script twice.
        if not requirements and graph_options.resume and project_info.runner.has_buildscript_call(project_info.script):
            with BuildscriptMetadata.capture() as future:
                project_info.runner.execute_script(project_info.script, {})
            assert future.done()
            requirements = RequirementSpec.from_metadata(future.result())

        # Update `sys.path` with the python path from the requirement spec, if any.
        if requirements:
            exit_stack.enter_context(appending_to_sys_path(requirements.pythonpath))

    context: Context | None = None

    # Deserialize the build state from files in the build state directory (+ extra dirs) if that is what
    # the user requested.
    if graph_options.resume:
        context, graph = serialize.load_build_state([build_options.state_dir] + build_options.additional_state_dirs)
        if not graph:
            raise ValueError("cannot --resume without build state")
        if graph and graph_options.restart:
            graph.restart()
        assert context is not None

    # Otherwise, we need to execute the build script.
    else:
        if build_options.no_load_project:
            raise ValueError(
                "no existing build state was loaded; typically that would load the root project "
                "but --no-load-project was specified."
            )

        # Register a callback for when the buildscript calls the buildscript() method. Any requirements passed
        # to the function are already expected to have been handled with by the Kraken wrapper, but we need to
        # handle the additions to `sys.path` here.
        def _buildscript_metadata_callback(metadata: BuildscriptMetadata) -> None:
            requirements = RequirementSpec.from_metadata(metadata)
            exit_stack.enter_context(appending_to_sys_path(requirements.pythonpath))

        context = Context(build_options.build_dir)

        with BuildscriptMetadata.callback(_buildscript_metadata_callback):
            context.load_project(Path.cwd())
            context.finalize()
            graph = TaskGraph(context)

    assert graph is not None

    # Serialize the build graph, even on failure, at the end of the build.
    if not graph_options.no_save:
        exit_stack.callback(
            lambda: serialize.save_build_state(build_options.state_dir, build_options.state_name, not_none(graph))
        )

    # Find the project from which we'll resolve relative task references based on the original current working
    # directory relative to the project root directory.
    relative_address = Address(":" + ":".join(subproject_directory.parts))

    # Find the project that contains the current working directory.
    try:
        context.focus_project = context.get_project(relative_address)
        logger.info("project '%s' is the main project", context.focus_project.address)
    except ProjectNotFoundError as e:
        logger.info("project '%s' does not exist, setting context.focus_project = None", e.address)
        context.focus_project = None

    # Deselect all tasks.
    for task in graph.root.tasks():
        task.selected = False

    # Mark tasks that were explicitly selected on the command-line as such. Tasks may alter their behaviour
    # based on whether they were explicitly selected or not.
    if graph_options.tasks:
        selected = context.resolve_tasks(graph_options.tasks, relative_to=relative_address, set_selected=True)
        targets = selected
    else:
        targets = context.resolve_tasks(None, relative_to=relative_address)

    # Trim the graph down to the selected or default tasks.
    if not graph_options.all:
        graph = graph.trim(targets)

    return context, graph


def run(
    exit_stack: contextlib.ExitStack,
    build_options: BuildOptions,
    graph_options: GraphOptions,
    run_options: RunOptions,
) -> None:
    context, graph = _load_build_state(
        exit_stack=exit_stack,
        build_options=build_options,
        graph_options=graph_options,
    )

    # TODO(@NiklasRosenstein): Having the observer mark the excluded tasks as skipped is a hack.
    #       Can we just mark these tasks as skipped after the graph is loaded (e.g. right here?).
    context.observer = ColoredDefaultPrintingExecutorObserver(
        context.resolve_tasks(run_options.exclude_tasks),
        context.resolve_tasks(run_options.exclude_tasks_subgraph),
    )

    if run_options.skip_build:
        print("note: skipped build due to -s,--skip-build option.")
        sys.exit(0)
    else:
        if not graph:
            if run_options.allow_no_tasks:
                print("note: no tasks were selected (--allow-no-tasks)", "blue", file=sys.stderr)
                sys.exit(0)
            else:
                print("error: no tasks were selected", file=sys.stderr)
                sys.exit(1)

        try:
            context.execute(graph)
        except BuildError as exc:
            print()
            print("error:", exc, file=sys.stderr)
            sys.exit(1)


def ls(graph: TaskGraph) -> None:
    goal_tasks = set(graph.tasks(goals=True))
    all_tasks = set(graph.tasks())
    if not all_tasks:
        print("no tasks")
        sys.exit(1)
    longest_name = max(map(len, (t.path for t in all_tasks))) + 1

    print()
    print(colored("Tasks", "blue", attrs=["bold", "underline"]))
    print()

    width = get_terminal_width(120)

    def _print_task(task: Task) -> None:
        line = [task.path.ljust(longest_name)]
        remaining_width = width - len(line[0])
        if task in goal_tasks:
            line[0] = colored(line[0], "green")
        if task.default:
            line[0] = colored(line[0], attrs=["bold"])
        status = graph.get_status(task)
        if status is not None:
            line.append(f"[{status_to_text(status)}]")
            status_length = 2 + len(status_to_text(status, colored=False)) + 1
            remaining_width -= status_length
        description = task.get_description()
        if description:
            remaining_width -= 2
            if remaining_width <= 0:
                remaining_width = width
            for part in textwrap.wrap(
                description,
                remaining_width,
                subsequent_indent=(width - remaining_width) * " ",
            ):
                line.append(part)
                line.append("\n")
            line.pop()
        print("  " + " ".join(line))

    def sort_key(task: Task) -> str:
        return task.path

    for task in sorted(graph.tasks(), key=sort_key):
        if isinstance(task, GroupTask):
            continue
        _print_task(task)

    print()
    print(colored("Groups", "blue", attrs=["bold", "underline"]))
    print()

    for task in sorted(graph.tasks(), key=sort_key):
        if not isinstance(task, GroupTask):
            continue
        _print_task(task)

    print()


def tree(graph: TaskGraph, exclude_options: ExcludeOptions) -> None:
    if exclude_options.exclude_tasks or exclude_options.exclude_tasks_subgraph:
        # Mark tasks that are excluded as skipped.
        for task in chain(
            graph.context.resolve_tasks(exclude_options.exclude_tasks),
            graph.context.resolve_tasks(exclude_options.exclude_tasks_subgraph),
        ):
            if not graph.get_status(task):
                graph.set_status(task, TaskStatus.skipped("excluded by -x/-X"))

    tasks = set(graph.tasks())

    # Filter out empty group tasks.
    tasks = {t for t in tasks if not isinstance(t, GroupTask) or t.tasks}

    # Find the projects we'd be looking to output.
    projects = {x.project for x in tasks}
    if graph.context.focus_project:
        projects.add(graph.context.focus_project)

    # Make sure we include all projects in between.
    for p in list(projects):
        while p and p.parent:
            p = p.parent
            projects.add(p)

    def _format_address(obj: Project | Task) -> str:
        address = obj.address

        parent = address.parent.set_container(True) if address and not address.is_root() else None
        if parent:
            result = colored(str(parent), "grey")
        else:
            result = ""
        result = result + colored(":" if address.is_root() else address.name, attrs=["bold"])

        if isinstance(obj, Project):
            if address.is_root():
                result += colored(" (root project)", "green")
            else:
                result += colored(" (sub project)", "green")
        if obj == graph.context.focus_project:
            result += colored(" (focus)", "cyan")
        if isinstance(obj, GroupTask):
            if obj.default:
                result += colored(" (default group)", "yellow", attrs=["bold"])
            else:
                result += colored(" (group)", "yellow")
        elif isinstance(obj, Task):
            if obj.default:
                result += colored(" (default task)", "blue", attrs=["bold"])
            else:
                result += colored(" (task)", "blue")
        if isinstance(obj, Task):
            if obj.selected:
                result += colored(" (selected)", "magenta")

            status = graph.get_status(obj)
            if status is not None:
                result += colored(f" [{status_to_text(status)}]", "cyan")

        return result

    def _recurse(obj: Project | Task, prefix: str, is_last: bool) -> None:
        if not (isinstance(obj, Project) and obj.address.is_root()):
            if is_last:
                this_prefix = prefix + "└── "
                child_prefix = prefix + "    "
            else:
                this_prefix = prefix + "├── "
                child_prefix = prefix + "│   "
        else:
            this_prefix = child_prefix = ""
        if isinstance(obj, Project):
            print(f"{colored(this_prefix, 'grey')}{_format_address(obj)}")
            members = sorted(list(obj.tasks().values()) + list(obj.subprojects().values()), key=lambda x: x.name)
            # Don't show empty groups
            members = [m for m in members if not isinstance(m, GroupTask) or m.tasks]
            # Don't show tasks/projects that are not in the graph.
            members = [m for m in members if m in tasks or m in projects]
            for idx, member in enumerate(members):
                is_last = idx == len(members) - 1
                _recurse(member, child_prefix, is_last)
        else:
            print(f"{colored(this_prefix, 'grey')}{_format_address(obj)}")

    print()
    _recurse(graph.context.root_project, "", True)


def describe(graph: TaskGraph) -> None:
    """
    Describes the selected tasks.
    """

    tasks = [t for t in graph.tasks() if t.selected]
    print("selected", len(tasks), "task(s)")
    print()

    for task in tasks:
        print("Group" if isinstance(task, GroupTask) else "Task", colored(task.path, attrs=["bold", "underline"]))
        print("  Type:", type(task).__module__ + "." + type(task).__name__)
        print("  Type defined in:", colored(sys.modules[type(task).__module__].__file__ or "???", "cyan"))
        print("  Default:", task.default)
        print("  Selected:", task.selected)
        print("  Capture:", task.capture)
        rels = list(task.get_relationships())
        print(colored("  Relationships", attrs=["bold"]), f"({len(rels)})")
        for rel in rels:
            print(
                "".ljust(4),
                colored(rel.other_task.path, "blue"),
                f"before={rel.inverse}, strict={rel.strict}",
            )
        print("  " + colored("Properties", attrs=["bold"]) + f" ({len(type(task).__schema__)})")
        longest_property_name = max(map(len, type(task).__schema__.keys())) if type(task).__schema__ else 0
        for key in type(task).__schema__:
            prop: Property[Any] = getattr(task, key)
            print(
                "".ljust(4),
                (key + ":").ljust(longest_property_name + 1),
                f'{colored(prop.get_or("<unset>"), "blue")}',
            )
        print()


def visualize(graph: TaskGraph, viz_options: VizOptions) -> None:
    root = graph.root
    if viz_options.reduce or viz_options.reduce_keep_explicit:
        root = root.reduce(keep_explicit=viz_options.reduce_keep_explicit)
        graph = graph.reduce(keep_explicit=viz_options.reduce_keep_explicit)

    buffer = io.StringIO()
    writer = GraphvizWriter(buffer if viz_options.show else sys.stdout)
    writer.digraph(fontname="monospace", rankdir="LR")
    writer.set_node_style(style="filled", shape="box")

    style_default = {"penwidth": "3"}
    style_goal = {"fillcolor": "lawngreen"}
    style_select = {"fillcolor": "darkgoldenrod1"}
    style_group = {"shape": "ellipse"}
    style_edge_non_strict = {"style": "dashed"}
    style_edge_implicit = {"color": "gray"}

    writer.subgraph("cluster_#legend", label="Legend")
    # writer.node("#task", label="task")
    writer.node("#group", label="group task", **style_group)
    writer.node("#default", label="runs by default", **style_default)
    writer.node("#selected", label="task will run", **style_select)
    writer.node("#goal", label="goal task", **style_goal)
    writer.end()

    writer.subgraph("cluster_#build", label="Build Graph")

    main = root if viz_options.inactive else graph
    goal_tasks = set(graph.tasks(goals=True))
    selected_tasks = set(graph.tasks())

    for task in main.tasks():
        style = {}
        style.update(style_default if task.default else {})
        style.update(style_group if isinstance(task, GroupTask) else {})
        style.update(style_select if task in selected_tasks else {})
        style.update(style_goal if task in goal_tasks else {})

        writer.node(task.path, **style)
        for predecessor in main.get_predecessors(task, ignore_groups=False):
            writer.edge(
                predecessor.path,
                task.path,
                **({} if main.get_edge(predecessor, task).strict else style_edge_non_strict),
                **(style_edge_implicit if main.get_edge(predecessor, task).implicit else {}),
            )

    writer.end()
    writer.end()

    if viz_options.show:
        render_to_browser(buffer.getvalue())


def env() -> None:
    dists = sorted(get_distributions().values(), key=lambda dist: dist.name)
    print(json.dumps([dist.to_json() for dist in dists], sort_keys=True))


def main_internal(prog: str, argv: list[str] | None, pdb_enabled: bool) -> NoReturn:
    parser = _get_argument_parser(prog)
    args = parser.parse_args(sys.argv[1:] if argv is None else argv)

    if not args.cmd:
        parser.print_usage()
        sys.exit(0)

    if LoggingOptions.available(args):
        LoggingOptions.collect(args).init_logging()

    if pdb_enabled:
        logger.info("note: KRAKEN_PDB=1 is set, an interactive debugging session will be started on exit.")
        logger.info("note: Exceptions raised in tasks will not trigger an interactive debugging session.")

    if args.cmd in ("run", "r"):
        with contextlib.ExitStack() as exit_stack:
            run(exit_stack, BuildOptions.collect(args), GraphOptions.collect(args), RunOptions.collect(args))

    elif args.cmd in ("query", "q"):
        if not args.query_cmd:
            parser.print_usage()
            sys.exit(0)

        if args.query_cmd == "env":
            env()
            sys.exit(0)

        build_options = BuildOptions.collect(args)
        graph_options = GraphOptions.collect(args)

        with contextlib.ExitStack() as exit_stack:
            _context, graph = _load_build_state(
                exit_stack=exit_stack,
                build_options=build_options,
                graph_options=graph_options,
            )

            if args.query_cmd == "ls":
                ls(graph)
            elif args.query_cmd in ("describe", "d"):
                describe(graph)
            elif args.query_cmd in ("visualize", "viz", "v"):
                visualize(graph, VizOptions.collect(args))
            elif args.query_cmd in ("t", "tree"):
                tree(graph, ExcludeOptions.collect(args))
            else:
                assert False, args.query_cmd

    else:
        parser.print_usage()

    sys.exit(0)


def main(prog: str = "kraken", argv: list[str] | None = None) -> NoReturn:
    pdb_enabled = os.getenv("KRAKEN_PDB") == "1"
    profile_outfile = os.getenv("KRAKEN_PROFILING")
    try:
        if profile_outfile:
            import cProfile as profile

            with open(profile_outfile, "w"):  # Make sure the file exists
                pass

            prof = profile.Profile()
            try:
                prof.runcall(main_internal, prog, argv, pdb_enabled)
            finally:
                prof.dump_stats(profile_outfile)
        else:
            main_internal(prog, argv, pdb_enabled)
        sys.exit(0)
    except:  # noqa: E722
        if pdb_enabled:
            pdb.post_mortem()
        raise


if __name__ == "__main__":
    main()



__version__ = "0.24.0"

from kraken.core.address import Address
from kraken.core.system.context import Context, ContextEvent
from kraken.core.system.errors import BuildError, ProjectLoaderError
from kraken.core.system.executor import Graph
from kraken.core.system.graph import TaskGraph
from kraken.core.system.project import Project
from kraken.core.system.property import Property
from kraken.core.system.task import (
    BackgroundTask,
    GroupTask,
    Task,
    TaskRelationship,
    TaskSet,
    TaskStatus,
    TaskStatusType,
    VoidTask,
)

__all__ = [
    "Address",
    "BackgroundTask",
    "BuildError",
    "Context",
    "ContextEvent",
    "Graph",
    "GroupTask",
    "Project",
    "ProjectLoaderError",
    "Property",
    "Task",
    "TaskGraph",
    "TaskRelationship",
    "TaskSet",
    "TaskStatus",
    "TaskStatusType",
    "VoidTask",
]



from __future__ import annotations

from pathlib import Path
from typing import Union

from kraken.common.path import try_relative_to
from kraken.common.supplier import Supplier

from kraken.core import Project, Property, Task, TaskStatus

from .check_file_contents_task import CheckFileContentsTask, as_bytes

DEFAULT_ENCODING = "utf-8"


class RenderFileTask(Task):
    """The RenderFileTask renders a single file to disk.

    The contents of the file can be provided by the :attr:`content` property or by creating a subclass
    that implements the :meth:`get_file_contents` method.

    It is common for a RenderFileTask to be added to the default `apply` task group. A matching check task,
    should you create it, would be a good candidate to add to the default `check` group.
    """

    description = 'Create or update "%(file)s".'

    file: Property[Path]
    content: Property[Union[str, bytes]]
    encoding: Property[str] = Property.default(DEFAULT_ENCODING)

    def create_check(
        self,
        name: str = "{name}.check",
        task_class: type[CheckFileContentsTask] | None = None,
        description: str | None = None,
        group: str | None = "check",
    ) -> CheckFileContentsTask:
        task = self.project.do(
            name.replace("{name}", self.name),
            task_class or CheckFileContentsTask,
            description=description,
            group=group,
            file=self.file.value,
            content=self.content.value,
            encoding=self.encoding.value,
            render_prepare=Supplier.of_callable(self.prepare),
        )
        task.add_relationship(self, strict=False)
        return task

    # Task

    def prepare(self) -> TaskStatus | None:
        from kraken.core.lib.check_file_contents_task import as_bytes

        file = self.file.get()
        if file.is_file() and file.read_bytes() == as_bytes(self.content.get(), self.encoding.get()):
            return TaskStatus.up_to_date(f'"{try_relative_to(file)}" is up to date')
        return TaskStatus.pending()

    def execute(self) -> TaskStatus:
        file = self.file.get()
        file.parent.mkdir(exist_ok=True, parents=True)
        content = as_bytes(self.content.get(), self.encoding.get())
        file.write_bytes(content)
        return TaskStatus.succeeded(f"wrote {len(content)} bytes to {try_relative_to(file)}")


def render_file(
    name: str,
    description: str | None = None,
    group: str | None = "apply",
    create_check: bool = True,
    check_name: str = "{name}.check",
    check_group: str | None = "check",
    check_description: str | None = None,
    project: Project | None = None,
    task_class: type[RenderFileTask] | None = None,
    check_task_class: type[CheckFileContentsTask] | None = None,
    *,
    file: str | Path | Supplier[Path],
    content: str | Supplier[str],
    encoding: str | Supplier[str] = DEFAULT_ENCODING,
) -> tuple[RenderFileTask, CheckFileContentsTask | None]:
    project = project or Project.current()
    render_task = project.do(
        name,
        task_class or RenderFileTask,
        description=description,
        group=group,
        file=file,
        content=content,
        encoding=encoding,
    )

    if create_check:
        check_task = render_task.create_check(
            check_name.replace("{name}", name),
            check_task_class,
            description=check_description,
            group=check_group,
        )
    else:
        check_task = None

    return render_task, check_task



from __future__ import annotations

import re
from dataclasses import dataclass
from typing import Any, ClassVar, Sequence, Tuple


class AddressMeta(type):
    """
    Meta class for #Address. Ensures that the copy-constructor returns the original address object.
    """

    def __call__(self, value: Any) -> Address:
        if isinstance(value, Address):
            return value
        obj = object.__new__(Address)
        obj.__init__(value)  # type: ignore[misc]
        return obj


class Address(metaclass=AddressMeta):
    """
    An address is an immutable parsed representation of a task or project reference, comparable to a filesystem path.
    The separator between elements in the address path is a colon (`:`). Similarly, a dot (`.`) refers to the current
    project, and a double dot (`..`) refers to the parent project.

    The elements of an address can only contain characters matching the :data:`Address.Element.VALIDATION_REGEX`.

    Asterisks are accepted to permit glob pattern matching on the addressable space, where one asterisk (`*`) is
    intended to match only within the same hierarchical level (aka. wildcard), wheras a double asterisk (`**`) is
    used to match any number of levels (aka. recursive wildcard). A trailing question mark on each element is allowed
    to permit that address resolution fails at that element.

    >>> Address(":a?:b").elements
    [Address.Element(value='a', fallible=True), Address.Element(value='b', fallible=False)]
    >>> Address("a:..:b").normalize()
    Address('b')
    """

    SEPARATOR: ClassVar[str] = ":"
    ROOT: ClassVar[Address]
    CURRENT: ClassVar[Address]
    PARENT: ClassVar[Address]
    EMPTY: ClassVar[Address]
    WILDCARD: ClassVar[Address]
    RECURSIVE_WILDCARD: ClassVar[Address]

    @dataclass(frozen=True)
    class Element:
        """Represents an element in the address, which is the text between colon separators (`:`)."""

        VALID_CHARACTERS: ClassVar[str] = r"a-zA-Z0-9/_\-\.\*"
        VALIDATION_REGEX: ClassVar[str] = rf"^[{VALID_CHARACTERS}]+$"
        CURRENT: ClassVar[str] = "."
        PARENT: ClassVar[str] = ".."
        RECURSIVE_WILDCARD: ClassVar[str] = "**"

        # The value of the element. This may contain asterisks for globbing.
        value: str

        #: Whether the element is followed by a question mark to permit resolution failure.
        fallible: bool = False

        def __post_init__(self) -> None:
            if not re.match(self.VALIDATION_REGEX, self.value):
                raise ValueError(f"invalid address element: {str(self)!r}")

        def __str__(self) -> str:
            if self.fallible:
                return f"{self.value}?"
            return self.value

        def is_current(self) -> bool:
            """
            Returns `True` if the element represents the current project (`.`).

            >>> Address.Element('.').is_current()
            True
            """

            return self.value == self.CURRENT

        def is_parent(self) -> bool:
            """
            Returns `True` if the element represents the parent project (`..`).

            >>> Address.Element('..').is_parent()
            True
            """

            return self.value == self.PARENT

        def is_concrete(self) -> bool:
            """
            Returns `True` if the element is a concrete element that can only have exactly one match. Elements that
            are not concrete have zero or more matches by asterisk (wildcards) in the #value or if #fallible is
            enabled.

            >>> Address.Element("test").is_concrete()
            True
            >>> Address.Element("test*").is_concrete()
            False
            >>> Address.Element("test", fallible=True).is_concrete()
            False
            """

            return not (self.fallible or "*" in self.value)

        def is_recursive_wildcard(self) -> bool:
            return self.value == self.RECURSIVE_WILDCARD

        @classmethod
        def of(cls, value: str) -> Address.Element:
            """
            Creates an element from a string. #fallible is set depending on whether *value* is trailed by a
            question mark.

            >>> Address.Element.of("test")
            Address.Element(value='test', fallible=False)
            >>> Address.Element.of("test?")
            Address.Element(value='test', fallible=True)
            >>> Address.Element.of("test??")
            Traceback (most recent call last):
            ValueError: invalid address element: 'test??'
            """

            fallible = False
            if value.endswith("?"):
                fallible = True
                value = value[:-1]
            return cls(value, fallible)

    _is_absolute: bool
    _is_container: bool
    _elements: list[Element]

    @staticmethod
    def _parse(value: str | Sequence[str]) -> tuple[bool, bool, list[Element]]:
        """Parses a list or strings or lists into (is_absolute, is_container, elements)."""

        # Convert the accepted types of value to a list of strings representing the elements of the address.
        is_absolute = False
        element_strings: list[str]
        if isinstance(value, str):
            if not value:
                element_strings = []
            elif value == Address.SEPARATOR:
                element_strings = [""]
            else:
                element_strings = value.split(Address.SEPARATOR)
        elif isinstance(value, Sequence):
            element_strings = list(value)
        else:
            assert False, type(value)

        # The first element may be an empty string to denote a "root" address.
        is_absolute = False
        if element_strings and not element_strings[0]:
            is_absolute = True
            element_strings.pop(0)

        # The last element may be an empty string to denote a "folder" address.
        is_container = False
        if element_strings and not element_strings[-1]:
            is_container = True
            element_strings.pop(-1)

        # Also, `:` is both absolute and a container
        if is_absolute and len(element_strings) == 0:
            is_container = True

        try:
            elements = [Address.Element.of(x) for x in element_strings]
        except ValueError as exc:
            raise ValueError(f"invalid Address: {Address.SEPARATOR.join(element_strings)!r} (reason: {exc})")

        if len(elements) == 0:
            # For some pathological addresses that are semantically equivalent to `:` (e.g. `:a:..`),
            # it is hard for the caller to easily detect this is a container.
            # Thus, we'll ensure it ourselves here.

            # The root object (":") is both absolute and a container
            # The empty address ("") is neither absolute nor a container
            if is_container or is_absolute:
                is_container = True
                is_absolute = True

        return is_absolute, is_container, elements

    @classmethod
    def create(cls, is_absolute: bool, is_container: bool, elements: list[Element]) -> Address:
        """
        Create a new address object.

        :param is_absolute: Whether the address is absolute (starts with `:`)
        :param elements: The address elements.

        >>> Address.create(True, False, [Address.Element("a", fallible=True), Address.Element("b")])
        Address(':a?:b')
        """

        obj = object.__new__(cls)
        obj._is_absolute = is_absolute
        obj._is_container = is_container
        obj._elements = elements
        obj._hash_key = None

        if len(elements) == 0:
            # For some pathological addresses that are semantically equivalent to `:` (e.g. `:a:..`),
            # it is hard for the caller to easily detect this is a container.
            # Thus, we'll ensure it ourselves here.

            # The root object (":") is both absolute and a container
            # The empty address ("") is neither absolute nor a container
            if obj._is_container or obj._is_absolute:
                obj._is_container = True
                obj._is_absolute = True

        return obj

    def __init__(self, value: str | Sequence[str] | Address) -> None:
        """Create a new Address from a string, sequence of strings or Address.

        >>> Address(":a:b")
        Address(':a:b')
        >>> Address(":a:b".split(":"))
        Address(':a:b')
        >>> Address(["", "a", "b"])
        Address(':a:b')

        Address objects are immutable and are not copied by the constructor (this is implemented via the
        meta class).

        >>> a = Address(':a')
        >>> a is Address(a)
        True

        Use #Address.create() to construct a new address object from a list of #Address.Element.
        """

        assert not isinstance(value, Address)
        self._is_absolute, self._is_container, self._elements = self._parse(value)
        self._hash_key: int | None = None

    def __getnewargs__(self) -> Tuple[Any, ...]:
        # For dill to serialize/deserialize the object.
        return (str(self),)

    def __str__(self) -> str:
        """
        Returns the string format of the address. Use the #Address constructor to parse it back into an address.

        >>> str(Address(":a:b"))
        ':a:b'
        """

        value = Address.SEPARATOR.join(str(x) for x in self._elements)
        if self._is_absolute:
            value = f"{Address.SEPARATOR}{value}"
        if self._is_container and not self.is_root():
            # The address must end with a separator...unless it is the root address `:`.
            value = f"{value}{Address.SEPARATOR}"
        return value

    def __repr__(self) -> str:
        """
        >>> repr(Address(":a?:b*"))
        "Address(':a?:b*')"
        """

        return f"Address({str(self)!r})"

    def __hash__(self) -> int:
        """
        Returns a stable hash key of the address.
        """

        if self._hash_key is None:
            self._hash_key = hash((Address, self._is_absolute, tuple(self._elements)))
        return self._hash_key

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Address):
            return (self._is_absolute, self._is_container, self._elements) == (
                other._is_absolute,
                other._is_container,
                other._elements,
            )
        return False

    def __len__(self) -> int:
        """
        Returns the number of elements in the address.

        >>> len(Address(":a:b:c"))
        3
        >>> len(Address("a:b:c"))
        3
        """

        return len(self._elements)

    def __getitem__(self, element_index: int) -> Address.Element:
        """
        Returns the _nth_ element in the address.

        >>> Address(":a:b")[1]
        Address.Element(value='b', fallible=False)
        """

        return self._elements[element_index]

    def is_empty(self) -> bool:
        """
        Returns `True` if the address is empty. The empty state is the only invalid state of an address.

        >>> Address("").is_empty()
        True
        >>> Address("a").is_empty()
        False
        >>> bool(Address(""))
        False
        >>> bool(Address("a"))
        True
        >>> Address.EMPTY == Address("")
        True
        """

        return not self._is_absolute and not self._elements

    def __bool__(self) -> bool:
        """Returns False if the address is empty, otherwise True.

        >>> bool(Address(":a:b"))
        True
        >>> bool(Address(""))
        False
        """

        return not self.is_empty()

    def is_absolute(self) -> bool:
        """
        Returns `True` if the address is absolute.

        >>> Address(":a").is_absolute()
        True
        >>> Address("a").is_absolute()
        False
        >>> Address("").is_absolute()
        False
        """

        return self._is_absolute

    def is_root(self) -> bool:
        """
        Returns `True` if the address is the root address (`:`).

        >>> Address(":").is_root()
        True
        >>> Address(":a").is_root()
        False
        >>> Address("a").is_root()
        False
        >>> Address("").is_root()
        False
        """
        return self._is_absolute and not self._elements

    def is_concrete(self) -> bool:
        """
        Returns `True` if this is a concrete address. A concrete address is one that is absolute and
        has no globbing elements (see #Address.Element.is_globbing()).

        >>> Address(":a:b").is_concrete()
        True
        >>> Address("a:b").is_concrete()
        False
        >>> Address(":*:b").is_concrete()
        False
        >>> Address(":a:b?").is_concrete()
        False
        """

        return self._is_absolute and all(x.is_concrete() for x in self._elements)

    def is_container(self) -> bool:
        """
        Returns `True` if this is a container address, that is, if it ends with a separator.

        >>> Address(":a:b").is_container()
        False
        >>> Address(":a:b:").is_container()
        True
        """

        return self._is_container

    def normalize(self) -> "Address":
        """
        Normalize the address, removing any superfluous elements (`.` for current, `..` for parent). A normalized
        is not a container address. Use #set_container() after #normalize() to make it a container address.

        >>> Address(":a:.:b").normalize()
        Address(':a:b')
        >>> Address(":a:..:b").normalize()
        Address(':b')
        >>> Address("..:.:b").normalize()
        Address('..:b')
        >>> Address("a:b:").normalize()
        Address('a:b')
        """

        elements: list[Address.Element] = []
        stack = list(reversed(self._elements))
        while stack:
            current = stack.pop()
            if current.is_parent() and elements:
                elements.pop()
            elif current.is_current():
                pass
            else:
                elements.append(current)
        if not self._is_absolute and not elements:
            elements = [Address.Element(Address.Element.CURRENT, False)]
        return Address.create(self._is_absolute, False, elements)

    def concat(self, address: "str | Address") -> "Address":
        """
        Concatenate two addresses. If *address* is absolute, return *address*.

        >>> Address(":a").concat("b:c")
        Address(':a:b:c')
        >>> Address(":a").concat(Address(":b"))
        Address(':b')
        """

        if isinstance(address, str):
            address = Address(address)
        if address._is_absolute:
            return address
        return Address.create(self._is_absolute, address._is_container, self._elements + address._elements)

    def append(self, element: str | Element) -> "Address":
        """
        Return a new address with one element appended.

        >>> Address(":").append("a")
        Address(':a')
        """

        if isinstance(element, str):
            element = Address.Element.of(element)
        assert isinstance(element, Address.Element), type(element)
        return Address.create(self._is_absolute, False, self._elements + [element])

    def set_container(self, is_container: bool) -> "Address":
        """
        Return a copy of this address with the container flag set to the given value. The container flag indicates
        whether the string representation of the address is followed by a colon (`:`). This status is irrelevant
        for the root address, as it is always a container.

        >>> Address(":a").set_container(True)
        Address(':a:')
        >>> Address(":a:").set_container(False)
        Address(':a')

        Attempting to set the container status to `False` for the root address will raise a #ValueError. Attempting
        to set any container status to the empty address will also raise a #ValueError.

        >>> Address(":").set_container(True)
        Address(':')
        >>> Address(":").set_container(False)
        Traceback (most recent call last):
        ValueError: Cannot set container status to False for root address
        >>> Address("").set_container(True)
        Traceback (most recent call last):
        ValueError: Cannot set container status for empty address
        """

        if self.is_root():
            if not is_container:
                raise ValueError("Cannot set container status to False for root address")
            return self
        if self.is_empty():
            raise ValueError("Cannot set container status for empty address")
        return Address.create(self._is_absolute, is_container, self._elements)

    @property
    def name(self) -> str:
        """
        Returns the value of the last element in the Address. If the address has no elements, which is
        the case for the root address or an empty address, a #ValueError will be raised.

        >>> Address(":a:b").name
        'b'
        >>> Address("a:b?").name
        'b'
        >>> Address(":").name
        Traceback (most recent call last):
        ValueError: Address(':') has no elements, and thus no name
        >>> Address("").name
        Traceback (most recent call last):
        ValueError: Address('') has no elements, and thus no name
        """

        if not self._elements:
            raise ValueError(f"{self!r} has no elements, and thus no name")
        return self._elements[-1].value

    @property
    def elements(self) -> list[Element]:
        """
        Returns the individual elements of the address. Note that you should also check #is_absolute() to
        understand whether the elements are to be interpreted relative or absolute.

        >>> Address(":").elements
        []
        >>> Address(":a:b").elements
        [Address.Element(value='a', fallible=False), Address.Element(value='b', fallible=False)]
        >>> Address(":a:b").elements == Address("a:b").elements
        True
        """

        return self._elements

    @property
    def parent(self) -> "Address":
        """
        Returns the parent address.

        >>> Address(":a:b").parent
        Address(':a')
        >>> Address(":a").parent
        Address(':')
        >>> Address("a").parent
        Address('.')
        >>> Address(".").parent
        Address('..')
        >>> Address("..").parent
        Address('..:..')

        The container status of the address is perserved.

        >>> Address(":a:b").parent
        Address(':a')
        >>> Address(":a:b:").parent
        Address(':a:')

        Use the #set_container() method to change the container status.

        The root and empty address have no parent.

        >>> Address(":").parent
        Traceback (most recent call last):
        ValueError: Root address has no parent
        >>> Address("").parent
        Traceback (most recent call last):
        ValueError: Empty address has no parent
        """

        if self._is_absolute and not self._elements:
            raise ValueError("Root address has no parent")
        if not self._is_absolute and not self._elements:
            raise ValueError("Empty address has no parent")

        # When we currently have a relative address '.' we want to return '..'.
        if not self._is_absolute and self._elements and self._elements[-1].is_current():
            return Address.create(False, self._is_container, [Address.Element(Address.Element.PARENT, False)])

        # When we currently have a relative address '..' we want to return '..:..'
        if not self._is_absolute and self._elements and self._elements[-1].is_parent():
            return Address.create(
                False, self._is_container, self._elements + [Address.Element(Address.Element.PARENT, False)]
            )

        if not self._is_absolute and len(self._elements) == 1:
            return Address.CURRENT

        # Strip the last element.
        assert self._elements, self
        return Address.create(self._is_absolute, self._is_container, self._elements[:-1])


Address.ROOT = Address(":")
Address.EMPTY = Address("")
Address.CURRENT = Address(".")
Address.PARENT = Address("..")
Address.WILDCARD = Address("*")
Address.RECURSIVE_WILDCARD = Address("**")



from __future__ import annotations

import dataclasses
import logging
from typing import TYPE_CHECKING, Iterable, Iterator, List, Sequence, TypeVar, cast

from kraken.common import not_none
from networkx import DiGraph, restricted_view, transitive_reduction
from networkx.algorithms import topological_sort
from nr.stream import Stream

from kraken.core.system.executor import Graph
from kraken.core.system.task import GroupTask, Task, TaskStatus

if TYPE_CHECKING:
    from kraken.core.system.context import Context

T = TypeVar("T")
logger = logging.getLogger(__name__)


@dataclasses.dataclass
class _Edge:
    strict: bool
    implicit: bool


class TaskGraph(Graph):
    """The task graph represents a Kraken context's tasks as a directed acyclic graph data structure.

    Before a task graph is passed to an executor, it is usually trimmed to contain only the tasks that are
    needed for the successful and complete execution of the desired set of "goal tasks"."""

    def __init__(self, context: Context, populate: bool = True, parent: TaskGraph | None = None) -> None:
        """Create a new build graph from the given task list.

        :param context: The context that the graph belongs to.
        :param populate: If enabled, the task graph will be immediately populated with the tasks in the context.
            The graph can also be later populated with the :meth:`populate` method.
        """

        self._parent = parent
        self._context = context

        # Nodes have the form {'data': _Node} and edges have the form {'data': _Edge}.
        self._digraph = DiGraph()

        # Keep track of task execution results.
        self._results: dict[str, TaskStatus] = {}

        # All tasks that have a successful or skipped status are stored here.
        self._ok_tasks: set[str] = set()

        # All tasks that have a failed status are stored here.
        self._failed_tasks: set[str] = set()

        # Keep track of the tasks that returned TaskStatus.STARTED. That means the task is a background task, and
        # if the TaskGraph is deserialized from a state file to continue the build, background tasks need to be
        # reset so they start again if another task requires them.
        self._background_tasks: set[str] = set()

        if populate:
            self.populate()

    def __bool__(self) -> bool:
        return len(self._digraph.nodes) > 0

    def __len__(self) -> int:
        return len(self._digraph.nodes)

    # Low level internal API

    def _get_task(self, task_path: str) -> Task | None:
        data = self._digraph.nodes.get(task_path)
        if data is None:
            return None
        try:
            return cast(Task, data["data"])
        except KeyError:
            raise RuntimeError(f"An unexpected error occurred when fetching the task by address {task_path!r}.")

    def _add_task(self, task: Task) -> None:
        self._digraph.add_node(task.path, data=task)
        for rel in task.get_relationships():
            if rel.other_task.path not in self._digraph.nodes:
                self._add_task(rel.other_task)
            a, b = (task, rel.other_task) if rel.inverse else (rel.other_task, task)
            self._add_edge(a.path, b.path, rel.strict, False)

            # If this relationship is one implied through group membership, we're done.
            if isinstance(task, GroupTask) and not rel.inverse and rel.other_task in task.tasks:
                continue

            # When a group depends on some other task, we implicitly make each member of that downstream group
            # depend on the upstream task. If we find another group, we unpack the group further.
            upstream, downstream = (task, rel.other_task) if rel.inverse else (rel.other_task, task)
            if isinstance(downstream, GroupTask):
                downstream_tasks = list(downstream.tasks)
                while downstream_tasks:
                    member = downstream_tasks.pop(0)
                    if member.path not in self._digraph.nodes:
                        self._add_task(member)
                    if isinstance(member, GroupTask):
                        downstream_tasks += member.tasks
                        continue

                    # NOTE(niklas.rosenstein): When a group is nested in another group, we would end up declaring
                    #       that the group depends on itself. That's obviously not supposed to happen. :)
                    if upstream != member:
                        self._add_edge(upstream.path, member.path, rel.strict, True)

    def _get_edge(self, task_a: str, task_b: str) -> _Edge | None:
        data = self._digraph.edges.get((task_a, task_b)) or self._digraph.edges.get((task_a, task_b))
        if data is None:
            return None
        return cast(_Edge, data["data"])

    def _add_edge(self, task_a: str, task_b: str, strict: bool, implicit: bool) -> None:
        # add_edge() would implicitly add a node, we only want to do that once the node actually exists in
        # the graph though.
        assert task_a in self._digraph.nodes, f"{task_a!r} not yet in the graph"
        assert task_b in self._digraph.nodes, f"{task_b!r} not yet in the graph"
        edge = self._get_edge(task_a, task_b) or _Edge(strict, implicit)
        edge.strict = edge.strict or strict
        edge.implicit = edge.implicit and implicit
        self._digraph.add_edge(task_a, task_b, data=edge)

    # High level internal API

    def _get_required_tasks(self, goals: Iterable[Task]) -> set[str]:
        """Internal. Return the set of tasks that are required transitively from the goal tasks."""

        def _is_empty_group_subtree(task_path: str) -> bool:
            """
            Returns `True` if the task pointed to by *task_path* is a GroupTask and it is empty or only depends on
            other empty groups.
            """

            def _is_empty_group(task_path: str) -> bool:
                """Returns `True` if the task pointed to by *task_path* is a GroupTask and it is empty."""

                task = self._get_task(task_path)
                if not isinstance(task, GroupTask):
                    return False
                return len(task.tasks) == 0

            def _is_empty_group_or_subtree(task_path: str) -> bool:
                """Returns `True` if the task pointed to by *task_path* is a GroupTask and it is empty or only depends
                on other empty groups."""

                task = self._get_task(task_path)
                if not isinstance(task, GroupTask):
                    return False
                for pred in self._digraph.predecessors(task_path):
                    if not _is_empty_group_or_subtree(pred):
                        return False
                return True

            return _is_empty_group(task_path) or _is_empty_group_or_subtree(task_path)

        def _recurse_task(task_path: str, visited: set[str], path: list[str]) -> None:
            if task_path in path:
                raise RuntimeError(f"encountered a dependency cycle: {' → '.join(path)}")
            visited.add(task_path)
            for pred in self._digraph.predecessors(task_path):
                if not_none(self._get_edge(pred, task_path)).strict:
                    # If the thing we want to pick up is a GroupTask and it doesn't have any members or other
                    # dependencies that are not also empty groups, we can skip it. It really doesn't need to be in the
                    # build graph.
                    if isinstance(self._get_task(pred), GroupTask):
                        # Check if the group is empty or only depends on other empty groups.
                        if _is_empty_group_subtree(pred):
                            continue
                    _recurse_task(pred, visited, path + [task_path])

        active_tasks: set[str] = set()
        for task in goals:
            _recurse_task(task.path, active_tasks, [])

        return active_tasks

    def _remove_nodes_keep_transitive_edges(self, nodes: Iterable[str]) -> None:
        """Internal. Remove nodes from the graph, but ensure that transitive dependencies are kept in tact."""

        for task_path in nodes:
            for in_task_path in self._digraph.predecessors(task_path):
                in_edge = not_none(self._get_edge(in_task_path, task_path))
                for out_task_path in self._digraph.successors(task_path):
                    out_edge = not_none(self._get_edge(task_path, out_task_path))
                    self._add_edge(
                        in_task_path,
                        out_task_path,
                        strict=in_edge.strict or out_edge.strict,
                        implicit=in_edge.implicit and out_edge.implicit,
                    )
            self._digraph.remove_node(task_path)

    def _get_ready_graph(self) -> DiGraph:
        """Updates the ready graph. Remove all ok tasks (successful or skipped) and any non-strict dependencies
        (edges) on failed tasks."""

        removable_edges: set[tuple[str, str]] = set()

        def set_non_strict_edge_for_removal(u: str, v: str) -> None:
            out_edge = not_none(self._get_edge(u, v))
            if not out_edge.strict:
                removable_edges.add((u, v))

        for failed_task_path in self._failed_tasks:
            for out_task_path in self._digraph.successors(failed_task_path):
                out_task = self._digraph.nodes[out_task_path]["data"]

                if isinstance(out_task, GroupTask):
                    # If the successor is a group task, check that the all of the groups tasks are either successful
                    # or failed, and then remove any non strict dependency (edge) on said group task.
                    group_task_paths = {task.path for task in out_task.tasks}
                    if not group_task_paths.issubset(self._failed_tasks | self._ok_tasks):
                        continue

                    for group_successor_path in self._digraph.successors(out_task_path):
                        set_non_strict_edge_for_removal(out_task_path, group_successor_path)
                else:
                    set_non_strict_edge_for_removal(failed_task_path, out_task_path)

        return restricted_view(self._digraph, self._ok_tasks, removable_edges)

    # Public API

    @property
    def context(self) -> Context:
        return self._context

    @property
    def parent(self) -> TaskGraph | None:
        return self._parent

    @property
    def root(self) -> TaskGraph:
        if self._parent:
            return self._parent.root
        return self

    def get_edge(self, pred: Task, succ: Task) -> _Edge:
        return not_none(self._get_edge(pred.path, succ.path), f"edge does not exist ({pred.path} --> {succ.path})")

    def get_predecessors(self, task: Task, ignore_groups: bool = False) -> List[Task]:
        """Returns the predecessors of the task in the original full build graph."""

        result = []
        for task in (not_none(self._get_task(task_path)) for task_path in self._digraph.predecessors(task.path)):
            if ignore_groups and isinstance(task, GroupTask):
                result += task.tasks
            else:
                result.append(task)
        return result

    def get_status(self, task: Task) -> TaskStatus | None:
        """Return the status of a task."""

        return self._results.get(task.path)

    def populate(self, goals: Iterable[Task] | None = None) -> None:
        """Populate the graph with the tasks from the context. This need only be called if the graph was
        not initially populated in the constructor.

        !!! warning "Inverse relationships"

            This does not recognize inverse relationships from tasks that are not part of *goals* or
            any of their relationships. It is therefore recommended to populate the graph with all tasks in the
            context and use #trim() to reduce the graph.
        """

        if goals is None:
            for project in self.context.iter_projects():
                for task in project.tasks().values():
                    if task.path not in self._digraph.nodes:
                        self._add_task(task)
        else:
            for task in goals:
                if task.path not in self._digraph.nodes:
                    self._add_task(task)

    def trim(self, goals: Sequence[Task]) -> TaskGraph:
        """Returns a copy of the graph that is trimmed to execute only *goals* and their strict dependencies."""

        graph = TaskGraph(self.context, parent=self)
        unrequired_tasks = set(graph._digraph.nodes) - graph._get_required_tasks(goals)
        graph._remove_nodes_keep_transitive_edges(unrequired_tasks)
        graph.results_from(self)
        return graph

    def reduce(self, keep_explicit: bool = False) -> TaskGraph:
        """Return a copy of the task graph that has been transitively reduced.

        :param keep_explicit: Keep non-implicit edges in tact."""

        digraph = self._digraph
        reduced_graph = transitive_reduction(digraph)
        reduced_graph.add_nodes_from(digraph.nodes(data=True))
        reduced_graph.add_edges_from(
            (u, v, digraph.edges[u, v])
            for u, v in digraph.edges
            if (keep_explicit and not digraph.edges[u, v]["data"].implicit) or (u, v) in reduced_graph.edges
        )

        graph = TaskGraph(self.context, populate=False, parent=self)
        graph._digraph = reduced_graph
        graph.results_from(self)

        return graph

    def results_from(self, other: TaskGraph) -> None:
        """Merge the results from the *other* graph into this graph. Only takes the results of tasks that are
        known to the graph. If the same task has a result in both graphs, and one task result is not successful,
        the not successful result is preferred."""

        self._results = {**other._results, **self._results}
        self._ok_tasks.update(other._ok_tasks)
        self._failed_tasks.update(other._failed_tasks)

        for task in self.tasks():
            status_a = self._results.get(task.path)
            status_b = other._results.get(task.path)
            if status_a is not None and status_b is not None and status_a.type != status_b.type:
                resolved_status: TaskStatus | None = status_a if status_a.is_not_ok() else status_b
            else:
                resolved_status = status_a or status_b
            if resolved_status is not None:
                # NOTE: This will already take care of updating :attr:`_background_tasks`.
                self.set_status(task, resolved_status, _force=True)

    def resume(self) -> None:
        """Reset the result of all background tasks that are required by any pending tasks. This needs to be
        called when a build graph is resumed in a secondary execution to ensure that background tasks are active
        for the tasks that require them."""

        reset_tasks: set[str] = set()
        for task in self.tasks(pending=True):
            for pred in self.get_predecessors(task, ignore_groups=True):
                if pred.path in self._background_tasks:
                    self._background_tasks.discard(pred.path)
                    self._ok_tasks.discard(pred.path)
                    self._failed_tasks.discard(pred.path)
                    self._results.pop(pred.path, None)
                    reset_tasks.add(pred.path)

        if reset_tasks:
            logger.info("Reset the status of %d background task(s): %s", len(reset_tasks), " ".join(reset_tasks))

    def restart(self) -> None:
        """Discard the results of all tasks."""

        self._results.clear()
        self._ok_tasks.clear()
        self._background_tasks.clear()
        self._failed_tasks.clear()

    def tasks(
        self,
        goals: bool = False,
        pending: bool = False,
        failed: bool = False,
        not_executed: bool = False,
    ) -> Iterator[Task]:
        """Returns the tasks in the graph in arbitrary order.

        :param goals: Return only goal tasks (i.e. leaf nodes).
        :param pending: Return only pending tasks.
        :param failed: Return only failed tasks.
        :param not_executed: Return only not executed tasks (i.e. downstream of failed tasks)"""

        tasks = (not_none(self._get_task(task_path)) for task_path in self._digraph)
        if goals:
            tasks = (t for t in tasks if self._digraph.out_degree(t.path) == 0)
        if pending:
            tasks = (t for t in tasks if t.path not in self._results)
        if failed:
            tasks = (t for t in tasks if t.path in self._results and self._results[t.path].is_failed())
        if not_executed:
            tasks = (
                t
                for t in tasks
                if (
                    (t.path not in self._results)
                    or (
                        t.path in self._results
                        and not self._results[t.path].is_succeeded()
                        and not self._results[t.path].is_failed()
                        and not self._results[t.path].is_skipped()
                        and not self._results[t.path].is_up_to_date()
                    )
                )
            )
        return tasks

    def execution_order(self, all: bool = False) -> Iterable[Task]:
        """Returns all tasks in the order they need to be executed.

        :param all: Return the execution order of all tasks, not just from the target subgraph."""

        order = topological_sort(self._digraph if all else self._get_ready_graph())
        return (not_none(self._get_task(task_path)) for task_path in order)

    # Graph

    def ready(self) -> list[Task]:
        """Returns all tasks that are ready to be executed. This can be used to constantly query the graph for new
        available tasks as the status of tasks in the graph is updated with :meth:`set_status`. An empty list is
        returned if no tasks are ready. At this point, if no tasks are currently running, :meth:`is_complete` can be
        used to check if the entire task graph was executed successfully."""

        ready_graph = self._get_ready_graph()
        root_set = (
            node for node in ready_graph.nodes if ready_graph.in_degree(node) == 0 and node not in self._results
        )
        tasks = [not_none(self._get_task(task_path)) for task_path in root_set]
        if not tasks:
            return []

        # NOTE(NiklasRosenstein): We don't need to return GroupTasks, we can mark them as skipped right away.
        #       In a future version of Kraken, we want to represent groups not as task objects, so this special
        #       handling code will be obsolete.
        result, groups = map(lambda x: list(x), Stream(tasks).bipartition(lambda t: isinstance(t, GroupTask)))
        for group in groups:
            self.set_status(group, TaskStatus.skipped())
        if not result:
            result = self.ready()
        return result

    def get_successors(self, task: Task, ignore_groups: bool = True) -> list[Task]:
        """Returns the successors of the task in the original full build graph.

        Never returns group tasks."""

        result = []
        for task in (not_none(self._get_task(task_path)) for task_path in self._digraph.successors(task.path)):
            if ignore_groups and isinstance(task, GroupTask):
                result += task.tasks
            else:
                result.append(task)
        return result

    def get_task(self, task_path: str) -> Task:
        if self._parent is None:
            return not_none(self._get_task(task_path))
        return self.root.get_task(task_path)

    def set_status(self, task: Task, status: TaskStatus, *, _force: bool = False) -> None:
        """Sets the status of a task, marking it as executed."""

        if not _force and (task.path in self._results and not self._results[task.path].is_started()):
            raise RuntimeError(f"already have a status for task {task.path!r}")
        self._results[task.path] = status
        if status.is_started():
            self._background_tasks.add(task.path)
        if status.is_ok():
            self._ok_tasks.add(task.path)
        if status.is_failed():
            self._failed_tasks.add(task.path)

    def is_complete(self) -> bool:
        """Returns `True` if, an only if, all tasks in the target subgraph have a non-failure result."""

        return set(self._digraph.nodes).issubset(self._ok_tasks)



""" Defines the Kraken executor API. """

from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Iterator

if TYPE_CHECKING:
    from kraken.core.system.task import Task, TaskStatus


class Graph(abc.ABC):
    """Interface for task graphs required for execution."""

    @abc.abstractmethod
    def ready(self) -> list[Task]:
        """Block until new tasks are ready to be executed. Return empty if no tasks are left. If no tasks are left
        but :meth:`is_complete` returns `False`, the build was unsuccessful."""

    @abc.abstractmethod
    def get_successors(self, task: Task) -> list[Task]:
        """Return all active dependants of the given task."""

    @abc.abstractmethod
    def get_task(self, task_path: str) -> Task:
        """Return a task by its path."""

    @abc.abstractmethod
    def set_status(self, task: Task, status: TaskStatus) -> None:
        """Set the result of a task. Can be called twice for the same task unless the previous call was passing
        a status with type :attr:`TaskStatusType.STARTED`."""

    @abc.abstractmethod
    def is_complete(self) -> bool:
        """Return `True` if all tasks in the graph are done and successful."""

    @abc.abstractmethod
    def tasks(
        self,
        goals: bool = False,
        pending: bool = False,
        failed: bool = False,
        not_executed: bool = False,
    ) -> Iterator[Task]:
        """Returns the tasks in the graph in arbitrary order.

        :param goals: Return only goal tasks (i.e. leaf nodes).
        :param pending: Return only pending tasks.
        :param failed: Return only failed tasks.
        :param not_executed: Return only not executed tasks (i.e. downstream of failed tasks)"""


class GraphExecutorObserver(abc.ABC):
    """Observes events in a Kraken task executor."""

    def before_execute_graph(self, graph: Graph) -> None:
        ...

    def before_prepare_task(self, task: Task) -> None:
        ...

    def after_prepare_task(self, task: Task, status: TaskStatus) -> None:
        ...

    def before_execute_task(self, task: Task, status: TaskStatus) -> None:
        ...

    def on_task_output(self, task: Task, chunk: bytes) -> None:
        ...

    def after_execute_task(self, task: Task, status: TaskStatus) -> None:
        ...

    def before_teardown_task(self, task: Task) -> None:
        ...

    def after_teardown_task(self, task: Task, status: TaskStatus) -> None:
        ...

    def after_execute_graph(self, graph: Graph) -> None:
        ...


class GraphExecutor(abc.ABC):
    @abc.abstractmethod
    def execute_graph(self, graph: Graph, observer: GraphExecutorObserver) -> None:
        ...



from pytest import raises

from kraken.core.system.project import Project
from kraken.core.system.property import Property
from kraken.core.system.task import Task, TaskRelationship


def test__Task__get_relationships_lineage_through_properties(kraken_project: Project) -> None:
    class MyTask(Task):
        prop: Property[str]

        def execute(self) -> None:
            raise NotImplementedError

    t1 = kraken_project.do("t1", MyTask)
    t1.prop.set("Hello, World")

    t2 = kraken_project.do("t2", MyTask)
    t2.prop.set(t1.prop)

    assert list(t2.get_relationships()) == [TaskRelationship(t1, True, False)]


def test__Task__new_style_type_hints_can_be_runtime_introspected_in_all_Python_versions(
    kraken_project: Project,
) -> None:
    """This is a feature of `typeapi ^1.3.0`."""

    class MyTask(Task):
        a: Property["list[str]"]  # type: ignore[misc]  # Mypy will complain in older Python versions even if this is not a runtime error  # noqa: E501
        b: Property["int | str"]

        def execute(self) -> None:
            raise NotImplementedError

    t1 = kraken_project.do("t1", MyTask)
    t1.a.set(["a", "b"])
    t1.b.set(42)
    t1.b.set("foo")

    with raises(TypeError) as excinfo:
        t1.a.set(("a", "b"))  # type: ignore[arg-type]
    assert str(excinfo.value) == "Property(MyTask(:t1).a): expected list, got tuple"
    with raises(TypeError) as excinfo:
        t1.b.set(42.0)  # type: ignore[arg-type]
    assert str(excinfo.value) == "Property(MyTask(:t1).b): expected int, got float\nexpected str, got float"



from __future__ import annotations

from pytest import raises

from kraken.core.address import Address


def test__Address__reuse() -> None:
    a1 = Address(":a:b")
    assert a1 is Address(a1)
    assert a1 is not Address(":a:b")


def test__Address__parse_empty_address() -> None:
    assert not Address("").is_absolute()
    assert Address("").elements == []
    assert str(Address("")) == ""


def test__Address__parse_root_address() -> None:
    assert Address(":").is_absolute()
    assert Address(":").elements == []
    assert Address(":").is_root()
    assert str(Address(":")) == ":"


def test__Address__parse_container_address() -> None:
    assert Address(":") == Address.ROOT
    assert Address(":").is_absolute()
    assert Address(":").is_container()

    assert Address(".") == Address.CURRENT
    assert not Address(".").is_absolute()
    assert not Address(".").is_container()
    assert Address(".:").is_container()

    # Absolute / Not Container
    assert Address(":a:b").is_absolute()
    assert not Address(":a:b").is_container()

    # Absolute / Container
    assert Address(":a:b:").is_absolute()
    assert Address(":a:b:").is_container()

    # Relative / Not Container
    assert not Address("a:b").is_absolute()
    assert not Address("a:b").is_container()

    # Relative / Container
    assert not Address("a:b:").is_absolute()
    assert Address("a:b:").is_container()

    assert Address(":a:..").normalize() == Address(":")
    assert Address("a:..:").normalize() == Address(".")


def test__Address__str() -> None:
    assert str(Address(":a")) == ":a"
    assert str(Address(":a:..:b")) == ":a:..:b"
    assert str(Address(":a?:c")) == ":a?:c"
    assert str(Address("a?:..")) == "a?:.."
    assert str(Address(":**:foo")) == ":**:foo"
    assert str(Address(":")) == ":"


def test__Address__cannot_contain_empty_element() -> None:
    with raises(ValueError):
        Address(":a::b")


def test__Address__cannot_contain_non_ascii_characters() -> None:
    with raises(ValueError):
        Address(":a:Ö")


def test__Address__concat() -> None:
    assert Address(".").concat("foo:bar") == Address(".:foo:bar")
    assert Address(":").concat("foo:bar") == Address(":foo:bar")
    assert Address("a:b").concat(":c") == Address(":c")

    assert str(Address(":a:").concat("b")) == ":a:b"
    assert str(Address(":a").concat("b:")) == ":a:b:"


def test__Address__normalize() -> None:
    assert Address("foo:..").normalize() == Address(".")
    assert Address("foo:bar:..").normalize() == Address("foo")
    assert Address(".:foo:bar").normalize() == Address("foo:bar")
    assert Address(".").normalize() == Address(".")
    assert Address("..").normalize() == Address("..")
    assert Address(":").normalize() == Address(":")
    assert Address(":..").normalize() == Address(":..")
    assert Address(":foo:bar:..").normalize() == Address(":foo")
    assert Address("b:..:a:..").normalize() == Address(".")
    assert Address(":b:..:a:..").normalize() == Address(":")
    assert Address(":**:a").normalize() == Address(":**:a")


def test__Address__parent_raises_ValueError_on_root() -> None:
    with raises(ValueError) as excinfo:
        Address(":").parent
    assert str(excinfo.value) == "Root address has no parent"


def test__Address__parent() -> None:
    assert Address(":a:b").parent == Address(":a")
    assert Address(":a:b").parent != Address(":a:")
    assert Address(".").parent == Address("..")
    assert Address("..").parent == Address("..:..")
    assert Address("..:a").parent == Address("..")
    assert Address("b:..:a").parent == Address("b:..")
    assert Address("b:..").parent == Address("b:..:..")


def test__Address__set_container() -> None:
    assert Address(":a:b").set_container(True) == Address(":a:b:")
    assert Address(":a:b:").set_container(True) == Address(":a:b:")
    assert Address(":a:b:").set_container(False) == Address(":a:b")
    assert Address(":a:..").set_container(False) == Address(":a:..")
    assert Address(":a:..:").set_container(True) == Address(":a:..:")
    assert Address(":a:..:").set_container(False) == Address(":a:..")



from typing_extensions import Protocol

from ._address import Address


class Addressable(Protocol):
    """
    A protocol that describes objects with an #address attribute. Such objects are considered living in an addressable
    space that can be resolved using an #_address_resolver.AddressSpace and #_address_resolver.resolve_address(). The
    address of any such object is expected to be concrete (see #Address.is_concrete()).
    """

    #: The concrete address of the object.
    address: Address
